@@ -11,14 +11,14 @@ from datetime import datetime
 from pathlib import Path
 from typing import List, TextIO, Dict, Tuple, Union, Iterator, Optional
 import multiprocessing as mp
-from tqdm import tqdm
 from nlcodec import __version__, log
 from nlcodec.dstruct import TrNode
-from nlcodec.utils import filter_types_coverage
+from nlcodec.utils import filter_types_coverage, IO
 import os
 import sys
 import random
 
+
 N_CPUS = max(1, mp.cpu_count() - 1)
 N_CPUS = int(os.environ.get('NLCODEC_THREADS', str(N_CPUS)))
 
@@ -58,10 +58,12 @@ class Reseved:
 class Level:
     reserved = -1
     char = 0
-    user = 0  # as of now,  user and char are indistinguishable; 0 means dont look inside
+    user = 0  # as of now,  user and char are indistinguishable; 0 means don't look inside
     subword = 1
     word = 2
     phrase = 3
+    clasz = 0   # 0 means dont split these tokens
+    byte = 0   # 0 means dont split these tokens
 
 
 @dataclass(frozen=True)
@@ -105,7 +107,7 @@ class Type:  # Type as in word type vs token
         return Type(**args)
 
     @classmethod
-    def write_out(cls, table: List['Type'], out: Union[Path, str, TextIO]):
+    def write_out(cls, table: List['Type'], out: Union[Path, str, TextIO], **meta):
 
         if isinstance(out, Path) or isinstance(out, str):
             wrtr = open(out, 'w', encoding='utf8', errors='ignore')
@@ -114,10 +116,11 @@ class Type:  # Type as in word type vs token
 
         levels = dict(coll.Counter(v.level for v in table))
         max_level = max(levels.keys())
-        meta = dict(total=len(table), version=__version__, levels=levels, max_level=max_level,
+        header = dict(total=len(table), version=__version__, levels=levels, max_level=max_level,
                     created=str(datetime.now()))
-        meta = json.dumps(meta)
-        wrtr.write(f"#{meta}\n")
+        header.update(meta)
+        header = json.dumps(header)
+        wrtr.write(f"#{header}\n")
         for i, item in enumerate(table):
             assert i == item.idx, f'{item} expected index {i}'
             wrtr.write(item.format() + '\n')
@@ -173,23 +176,20 @@ class Type:  # Type as in word type vs token
             return [self.name if name else self.idx]
 
 
-class EncoderScheme:
+class EncoderScheme(abc.ABC):
 
-    def __init__(self, table: List[Type], validate=True, invertible=True):
+    name = ''
+
+    def __init__(self, table: List[Type], has_reserved=True, invertible=True):
         """
 
         :param table: list of `Type`s
-        :param validate: validate that reserved types are found
+        :param has_reserved: validate that reserved types are found
         :param invertible: validate that the idx->str and str->idx are invertible
         """
-        if validate:
+        if has_reserved:
             Reseved.validate(table)
             self.unk_idx = Reseved.UNK_IDX
-        else:
-            # at least UNK should be available
-            assert table[Reseved.UNK_IDX].name == Reseved.UNK_TOK[0]
-            # TODO: reverse lookup UNK IDX based on UNK_TOK name
-            self.unk_idx = Reseved.UNK_IDX
 
         self.vocab_size = len(table)
         self.table = table
@@ -211,7 +211,6 @@ class EncoderScheme:
     def __len__(self):
         return self.vocab_size
 
-
     @abc.abstractmethod
     def encode_str(cls, line: str) -> List[str]:
         raise NotImplementedError()
@@ -287,6 +286,68 @@ class EncoderScheme:
         log.info(f"Total {cls} vocab size {len(vocab):,}")
         return vocab
 
+    def shrink_vocab(self, files: List[Path], min_freq: int, save_at: Optional[Path] = None) -> List[int]:
+        """
+        :param files:
+        :param min_freq:
+        :param save_at:
+        :return:
+        """
+        """"
+- Accept a list of files
+- compute term frequencies
+- Eliminate types with zero counts
+- Preserve reserved types even if they have zero counts
+- Save the resulting model at a given file path
+- Return index mapping between old and new, so we can go back to model and shrink embedding tables
+        """
+        from tqdm import tqdm
+        freqs = coll.Counter()
+        for file in files:
+            log.info(f'Computing term frequencies from {file}')
+            with IO.reader(file) as lines:
+                for line in tqdm(lines):
+                    freqs.update(self.encode(line))
+        assert len(self.table) > max(freqs.keys())
+        removals = [False] * len(self.table)
+        for idx, typ in enumerate(self.table):
+            if typ.level == Level.reserved:
+                continue # i.e. don't remove
+            removals[idx] = freqs[idx] < min_freq  # remove if min_freq threshold not met
+
+        # now make sure to preserve all the sub pieces leading to the pieces that retain
+        for idx in range(len(self.table) - 1, -1, -1):
+            combo = self.table[idx]
+            assert combo.idx == idx
+            if not removals[combo.idx] and combo.kids: #
+                for piece in combo.kids:
+                    if removals[piece.idx]:
+                        removals[piece.idx] = False   # dont remove this piece,
+
+        mapping = []
+        for idx, (is_remove, typ) in enumerate(zip(removals, self.table)):
+            assert idx == typ.idx
+            if is_remove:
+                continue
+            mapping.append(idx)
+
+        log.info(f"Shrinking vocab tables: {len(self.table)} --> {len(mapping)} ")
+        rev_mapping = {old_idx: new_idx for new_idx, old_idx in enumerate(mapping)}
+        old_table = self.table
+        new_table = []
+        for new_idx, old_idx in enumerate(mapping):
+            assert len(new_table) == new_idx
+            old_t = old_table[old_idx]
+            new_kids = [new_table[rev_mapping[k.idx]] for k in old_t.kids] if old_t.kids else None
+            new_t = old_t.copy(idx=new_idx, kids=new_kids)
+            new_table.append(new_t)
+        if save_at:
+            Type.write_out(new_table, out=save_at)
+        return mapping
+
+    def save(self, out):
+        return Type.write_out(table=self.table, out=out, scheme=self.name)
+
 
 class WordScheme(EncoderScheme):
     level = Level.word
@@ -302,6 +363,7 @@ class WordScheme(EncoderScheme):
 
     @classmethod
     def term_frequencies(cls, data: Iterator[str]) -> Tuple[Dict[str, int], int]:
+        from tqdm import tqdm
         stats = coll.Counter()
         line_count = 0
         for line in tqdm(data, mininterval=1):
@@ -392,7 +454,6 @@ class BPEScheme(CharScheme):
         assert not root.has_data  # root node is not data node
         return root
 
-
     def encode(self, line: str, split_ratio: float = 0.) -> List[int]:
         pieces = self.encode_str(line, split_ratio=split_ratio)
         return [self.str_to_idx.get(piece, self.unk_idx) for piece in pieces]
@@ -472,24 +533,124 @@ class BPEScheme(CharScheme):
         return res
 
 
+class ClassScheme(WordScheme):
+    """Scheme to be used for mapping labels or classes"""
+    level = Level.clasz
+    name = "class"
+    delim = ","
+    unk_idx = -1    # No unks
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, has_reserved=False, **kwargs)
+
+
+    @classmethod
+    def encode_str(cls, line: str) -> List[str]:
+        return [line.strip()]
+
+    def decode_str(self, seq: List[str]) -> str:
+        return self.delim.join(seq)
+
+    @classmethod
+    def get_init_vocab(cls, term_freqs, *args, **kwargs):
+        vocab = []
+        # sort alphabetically
+        term_freqs = sorted(term_freqs.items(), key=lambda x: x[0], reverse=False)
+
+        vocab += [Type(name=name, idx=idx, freq=freq, level=cls.level)
+                  for idx, (name, freq) in enumerate(term_freqs, start=len(vocab))]
+        log.info(f"Total {cls} vocab size {len(vocab):,}")
+        return vocab
+
+
+class ByteScheme(EncoderScheme):
+    level = Level.byte
+    name = "byte"
+    """
+     using hex strings to represent bytes [0-255] => [00-ff] 
+     <s> aka BOS is 256
+     </s> aka EOS is 257
+    """
+    def __init__(self, table: List[Type]=None, encoding='utf-8', errors="replace"):
+        self.encoding = encoding
+        self.errors = errors   # very likely, model is going to generate invalid code bytes during training
+        table = table or self.get_init_vocab()
+        super().__init__(table=table, has_reserved=False)
+
+
+    @staticmethod
+    def code_to_str(code: int) -> str:
+        return f'{code:x}'
+
+    def compose_str(self, pieces: List[str]):
+        byte_arr = bytes.fromhex(''.join(pieces))
+        return str(byte_arr, encoding=self.encoding, errors=self.errors)
+
+    def encode_str(self, line: str) -> List[str]:
+        return [self.code_to_str(b) for b in str.encode(line, self.encoding)]
+
+    def decode_str(self, seq: List[str]) -> str:
+        builder = []  # string builder
+        buffer = []   # buffer of past pieces
+        for piece in seq:
+            if piece in self.str_to_idx and self.table[self.str_to_idx[piece]].is_reserved:
+                if buffer:
+                    builder.append(self.compose_str(pieces=buffer))
+                    buffer.clear()
+                builder.append(piece)
+            else:
+                buffer.append(piece)
+        if buffer:
+            builder.append(self.compose_str(pieces=buffer))
+        return ''.join(builder)
+
+    def encode(self, line: str) -> List[int]:
+        pieces = self.encode_str(line)
+        return [self.str_to_idx[piece] for piece in pieces]
+
+    def decode(self, seq: List[int]) -> str:
+        pieces = [self.idx_to_str[idx] for idx in seq]
+        return self.decode_str(pieces)
+
+    @classmethod
+    def get_init_vocab(cls, *args, **kwargs):
+        vocab = [Type(name=f'{code:x}', idx=code, freq=-1, level=cls.level) for code in range(256)]
+        for tok, _ in [Reseved.BOS_TOK, Reseved.EOS_TOK]:
+            vocab.append(Type(name=tok, idx=len(vocab), freq=-1, level=Level.reserved))
+        log.info(f"Total {cls} vocab size {len(vocab):,}")
+        return vocab
+
+
+    @classmethod
+    def learn(cls, *args, **kwargs) -> List[Type]:
+        if args or kwargs:
+            log.warning(f"Byte vocabulary does not need learning; args are ignored: {args} {kwargs}")
+        return cls.get_init_vocab()
+
+
 #########################
 REGISTRY = {
     'char': CharScheme,
     'word': WordScheme,
     'bpe': BPEScheme,
-    'subword': BPEScheme
+    'subword': BPEScheme,
+    'class': ClassScheme,
+    'byte': ByteScheme
 }
 
 
 def learn_vocab(inp, level, model, vocab_size, min_freq=1, term_freqs=False,
-                char_coverage=CHAR_COVERAGE, min_co_ev=MIN_CO_EV):
+                char_coverage=CHAR_COVERAGE, min_co_ev=MIN_CO_EV) -> List[Type]:
     if not min_freq or min_freq < 1:
         min_freq = WORD_MIN_FREQ if level == 'word' else CHAR_MIN_FREQ
         log.info(f"level={level} => default min_freq={min_freq}")
     else:
         log.info(f"level={level} => user given min_freq={min_freq}")
     log.info(f"Learn Vocab for level={level} and store at {model}")
-    log.info(f"data ={inp}")
+    if isinstance(inp, (list, tuple)):
+        log.info(f"data ={inp[:10]} ... + ({len(inp) - 10} more items)")
+    else:
+        log.info(f"data ={inp}")
     Scheme = REGISTRY[level]
     args = {}
     if level != 'word':
@@ -498,16 +659,22 @@ def learn_vocab(inp, level, model, vocab_size, min_freq=1, term_freqs=False,
         args['min_co_evidence'] = min_co_ev
     table = Scheme.learn(inp, vocab_size=vocab_size, min_freq=min_freq, term_freqs=term_freqs,
                          **args)
-    Type.write_out(table=table, out=model)
+    if model:
+        Scheme(table).save(out=model)
+    return table
 
 
 def load_scheme(path: Union[str, Path, TextIO]) -> EncoderScheme:
     types, meta = Type.read_vocab(path)
     assert meta
-    max_level = meta['max_level']
-    levels = [CharScheme, BPEScheme, WordScheme]
-    assert max_level < len(levels)
-    Scheme = levels[max_level]
+    if 'scheme' in meta:
+        Scheme = REGISTRY[meta['scheme']]
+    else:
+        # backward compatibility;
+        max_level = meta['max_level']
+        levels = [CharScheme, BPEScheme, WordScheme, ClassScheme]
+        assert max_level < len(levels)
+        Scheme = levels[max_level]
     return Scheme(table=types)
 
 
